from pathlib import Path
import subprocess

base_dir = Path(__file__).parents[1]
pyproject_file = str(base_dir  / "pyproject.toml")
requirements_file = str(base_dir / "requirements.txt")
constraints_file = str(base_dir / "constraints.txt")

def upgrade_requirements():
    subprocess.run(
        [
            "uv", "pip", "compile",
            pyproject_file,
            "--universal",
            "--generate-hashes",
            "-o", requirements_file,
            "--constraint", constraints_file,
            "--upgrade",
        ]
    )

if __name__ == "__main__":
    upgrade_requirements()



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from unittest.mock import patch

from ducktools.pytui.shells import _core as core, powershell


POWERSHELL_PATH = R"C:\Program Files\PowerShell\7\pwsh.exe"
WINDOWS_POWERSHELL_PATH = R"C:\Windows\System32\WindowsPowerShell\v1.0\powershell.exe"


def test_registered():
    assert powershell.PowerShell.bin_name in core.Shell.registry
    assert powershell.WindowsPowerShell.bin_name in core.Shell.registry

    assert (powershell.PowerShell
            is core.Shell.registry[powershell.PowerShell.bin_name])
    assert (powershell.WindowsPowerShell
            is core.Shell.registry[powershell.WindowsPowerShell.bin_name])


def test_frompath():
    assert (core.Shell.from_path(POWERSHELL_PATH)
            == powershell.PowerShell(POWERSHELL_PATH))
    assert (core.Shell.from_path(WINDOWS_POWERSHELL_PATH)
            == powershell.WindowsPowerShell(WINDOWS_POWERSHELL_PATH))


def test_get_venv_shell_command():
    venv = R"C:\Users\David\Source\result\.venv"
    path = (
        R"C:\Users\David\Source\result\.venv\Scripts;"
        R"C:\Windows\system32;"
        R"C:\Users\David\.local\bin;"
        R"C:\Program Files\Github CLI"
    )
    prompt = "pytui: .venv"

    env = {
        "PYTUI_PATH": path,
        "PYTUI_VIRTUAL_ENV": venv,
        "PYTUI_VIRTUAL_ENV_PROMPT": prompt,
    }

    with patch("ducktools.pytui.shells.powershell.get_shell_script") as script_mock:
        def identity(x): return x

        script_mock.side_effect = identity

        shell = powershell.PowerShell(POWERSHELL_PATH)

        cmd, env_updates = shell.get_venv_shell_command(env)

        assert cmd == [POWERSHELL_PATH, "-NoExit", "activate_pytui.ps1"]
        assert env_updates == {}



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import json
import os
import os.path
import shutil
from typing import ClassVar

from ducktools.classbuilder.prefab import Prefab, as_dict, attribute

from .shells import Shell

from .platform_paths import (
    CONFIG_FILE, GLOBAL_VENV_FOLDER,
)


class Config(Prefab, kw_only=True):
    VENV_SEARCH_MODES: ClassVar[list[str]] = [
        "cwd", "parents", "recursive", "recursive_parents"
    ]

    config_file: str = attribute(default=CONFIG_FILE, serialize=False)
    venv_search_mode: str = "parents"
    include_pip: bool = True
    latest_pip: bool = True
    global_venv_folder: str = GLOBAL_VENV_FOLDER
    shell_path: str | None = None

    @property
    def shell(self) -> Shell | None:
        if self.shell_path is None:
            shell = Shell.get_default()
            if shell:
                self.shell_path = shell.path
                self.write_config()  # Save the updated shell_path
        else:
            shell = Shell.from_path(self.shell_path)

        return shell

    def set_shell(self, shell_path: str) -> str | None:
        if not os.path.isfile(shell_path):
            out_path = shutil.which(shell_path)
        else:
            out_path = shell_path

        if out_path and Shell.from_path(out_path) is not None:
            self.shell_path = out_path

        return out_path

    def write_config(self) -> None:
        os.makedirs(os.path.dirname(self.config_file), exist_ok=True)
        data = as_dict(self)

        with open(self.config_file, 'w') as f:
            json.dump(data, f, indent=4)

    @classmethod
    def from_file(cls, config_file=CONFIG_FILE) -> Config:
        if os.path.exists(config_file):
            with open(config_file, 'r') as f:
                try:
                    raw_input = json.load(f)
                except json.JSONDecodeError:
                    raw_input = {}

            venv_search_mode = raw_input.get("venv_search_mode", "parents")
            include_pip = raw_input.get("include_pip", True)
            latest_pip = raw_input.get("latest_pip", True)
            global_venv_folder = raw_input.get("global_venv_folder", GLOBAL_VENV_FOLDER)
            shell_path = raw_input.get("shell_path", None)

            if venv_search_mode not in cls.VENV_SEARCH_MODES:
                venv_search_mode = "parents"
            if not isinstance(include_pip, bool):
                include_pip = True
            if not isinstance(latest_pip, bool):
                latest_pip = True

            config = cls(
                config_file=config_file,
                venv_search_mode=venv_search_mode,
                include_pip=include_pip,
                latest_pip=latest_pip,
                global_venv_folder=global_venv_folder,
                shell_path=shell_path,
            )

            if raw_input != as_dict(config):
                config.write_config()

        else:
            config = cls(config_file=config_file)
            config.write_config()
        return config



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import sys
from functools import lru_cache

from ducktools.pythonfinder import PythonInstall

from .base import PythonListing, RuntimeManager

# Windows python installer should come before UV if available
if sys.platform == "win32":
    from . import pythoncore as pythoncore
    from . import uv as uv
else:
    from . import uv as uv


@lru_cache(maxsize=None)
def get_managers() -> list[RuntimeManager]:
    managers = []
    for m in RuntimeManager.available_managers:
        inst = m()
        if inst.executable:
            managers.append(inst)
    return managers


def fetch_downloads() -> list[PythonListing]:
    downloads = []
    for m in get_managers():
        downloads.extend(m.fetch_downloads())

    return downloads


def find_matching_listing(install: PythonInstall) -> PythonListing | None:
    for manager in get_managers():
        listing = manager.find_matching_listing(install)
        if listing:
            break
    else:
        listing = None

    return listing



# mypy: disable-error-code="attr-defined"
# attributes don't exist on non-windows platform
# but this entire file is only imported on windows.
"""
Helper code to 'fix' the windows terminal history in order to get command history
in Python that doesn't use PyREPL.

Copied from:
https://discuss.python.org/t/interactive-command-history-in-session-started-with-subprocess-on-windows/3701/5
"""
from __future__ import annotations

import ctypes
import collections
from ctypes import wintypes

kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)

HISTORY_NO_DUP_FLAG = 1


class ConsoleHistoryInfoStruct(ctypes.Structure):
    _fields_ = (
        ('cbSize', wintypes.UINT),
        ('HistoryBufferSize', wintypes.UINT),
        ('NumberOfHistoryBuffers', wintypes.UINT),
        ('dwFlags', wintypes.DWORD)
    )

    def __init__(self, *args, **kwds):
        super().__init__(ctypes.sizeof(self), *args, **kwds)


ConsoleHistoryInfo = collections.namedtuple(
    'ConsoleHistoryInfo',
    'bufsize nbuf flags'
)


def get_console_history_info():
    info = ConsoleHistoryInfoStruct()
    if not kernel32.GetConsoleHistoryInfo(ctypes.byref(info)):
        raise ctypes.WinError(ctypes.get_last_error())

    return ConsoleHistoryInfo(
        info.HistoryBufferSize,
        info.NumberOfHistoryBuffers,
        info.dwFlags
    )


def set_console_history_info(
    bufsize=512,
    nbuf=32,
    flags=HISTORY_NO_DUP_FLAG
):
    info = ConsoleHistoryInfoStruct(bufsize, nbuf, flags)
    if not kernel32.SetConsoleHistoryInfo(ctypes.byref(info)):
        raise ctypes.WinError(ctypes.get_last_error())



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from unittest.mock import patch

from ducktools.pytui.shells import _core as core, zsh


ZSH_PATH = "/bin/zsh"


def test_registered():
    assert zsh.ZShell.bin_name in core.Shell.registry
    assert zsh.ZShell is core.Shell.registry[zsh.ZShell.bin_name]


def test_frompath():
    assert core.Shell.from_path(ZSH_PATH) == zsh.ZShell(ZSH_PATH)


def test_get_venv_shell_command():
    venv = "/home/david/src/ducktools-pytui/.venv"
    path_env = (
        "/home/david/src/ducktools-pytui/.venv"
        "/home/david/.pyenv/shims:"
        "/home/david/.cargo/bin:"
        "/home/david/.pyenv/bin:"
        "/home/david/.local/bin:"
        "/home/david/bin:"
        "/usr/local/sbin:"
        "/usr/local/bin:"
        "/usr/sbin:"
        "/usr/bin:"
        "/sbin:"
        "/bin:"
        "/usr/games:"
        "/usr/local/games:"
        "/snap/bin"
    )
    prompt = "pytui: .venv"

    env = {
        "PYTUI_PATH": path_env,
        "PYTUI_VIRTUAL_ENV": venv,
        "PYTUI_VIRTUAL_ENV_PROMPT": prompt,
    }

    shell = zsh.ZShell(ZSH_PATH)

    cmd, env_updates = shell.get_venv_shell_command(env)

    assert cmd == [ZSH_PATH, "--no-rcs"]
    assert env_updates == {
        "PATH": path_env,
        "VIRTUAL_ENV": venv,
        "VIRTUAL_ENV_PROMPT": prompt,
        "PS1": "(pytui: .venv) %n@%m:%~/ > "
    }



# This file is a part of ducktools.pytui
# A TUI for managing Python installs and virtual environments
#
# Copyright (C) 2025  David C Ellis
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import functools
import operator
import os.path
import subprocess
import sys
from abc import ABC, abstractmethod
from collections.abc import Iterable
from typing import ClassVar

from ducktools.pythonfinder.shared import PythonInstall, version_str_to_tuple
from ducktools.classbuilder.prefab import prefab, attribute

# typing hack - avoid deprecated alias on newer python
if sys.version_info >= (3, 9):
    _version_tuple_type = tuple[int, int, int, str, int]
else:
    from typing import Tuple
    _version_tuple_type = Tuple[int, int, int, str, int]


class RuntimeManager(ABC):
    available_managers: ClassVar[list[type[RuntimeManager]]] = []
    organisation: ClassVar[str | None] = None

    def __init_subclass__(cls):
        RuntimeManager.available_managers.append(cls)

    @staticmethod
    def sort_listings(listings: Iterable[PythonListing]):
        new_listings = sorted(listings, key=operator.attrgetter("variant", "arch", "key"))
        new_listings.sort(key=operator.attrgetter("version_tuple"), reverse=True)
        new_listings.sort(key=operator.attrgetter("implementation"))
        return new_listings

    @functools.cached_property
    @abstractmethod
    def executable(self) -> str | None:
        """
        Get the path to the manager executable or None if it is not installed
        """
        ...

    @abstractmethod
    def fetch_installed(self) -> list[PythonListing]:
        """
        Get a list of installed runtimes managed by the manager
        """
        ...

    @abstractmethod
    def _get_download_cache(self):
        """
        List all available downloads (cached method)
        """

    @abstractmethod
    def fetch_downloads(self) -> list[PythonListing]:
        """
        List available downloads, exclude already downloaded (not cached)
        """

    def find_matching_listing(self, install: PythonInstall) -> PythonListing | None:
        if install.managed_by is None or not install.managed_by.startswith(self.organisation):
            return None

        # Executable names may not match, one may find python.exe, the other pypy.exe
        # Use the parent folder.
        installed_dict = {
            os.path.dirname(os.path.abspath(py.path)): py
            for py in self.fetch_installed()
            if py.path is not None
        }

        install_path = os.path.dirname(install.executable)

        return installed_dict.get(install_path, None)


@prefab(kw_only=True)
class PythonListing(ABC):
    manager: RuntimeManager

    key: str
    version: str
    implementation: str
    variant: str
    arch: str
    path: str | None
    url: str | None

    # `attribute` is a field specifier as defined in dataclass_transform
    # Not sure why it's not being picked up
    _version_tuple: _version_tuple_type | None = attribute(default=None, private=True)

    @property
    def version_tuple(self) -> _version_tuple_type:
        if not self._version_tuple:
            self._version_tuple = version_str_to_tuple(self.version)
        return self._version_tuple

    @property
    def full_key(self):
        return f"{type(self).__name__} / {self.key}"

    @property
    def will_overwrite(self):
        return False

    @classmethod
    @abstractmethod
    def from_dict(cls, manager, entry):
        ...

    @abstractmethod
    def install(self) -> subprocess.CompletedProcess | None:
        ...

    @abstractmethod
    def uninstall(self) -> subprocess.CompletedProcess | None:
        ...


# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

from unittest.mock import patch

from ducktools.pytui.shells import _core as core, bash

BASH_PATH = R"C:\Program Files\Git\usr\bin\bash.exe"


def test_registered():
    assert bash.GitBashShell.bin_name in core.Shell.registry
    assert bash.GitBashShell is core.Shell.registry[bash.GitBashShell.bin_name]


def test_frompath():
    assert core.Shell.from_path(BASH_PATH) == bash.GitBashShell(BASH_PATH)


def test_dedupe_path():
    venv_path = R"C:\Users\David\Source\ducktools-pytui\.venv\Scripts"
    bash_path = (
        "/c/Users/David/Source/ducktools-pytui/.venv/Scripts:"
        "/mingw64/bin:"
        "/usr/bin:"
        "/c/Users/David/.local/bin:"
        "/usr/bin:"
    )

    shell = bash.GitBashShell(BASH_PATH)

    expected = (
        "/c/Users/David/Source/ducktools-pytui/.venv/Scripts:"
        "/mingw64/bin:"
        "/usr/bin:"
        "/c/Users/David/.local/bin"
    )

    assert shell.get_deduped_path(bash_path, venv_path) == expected


def test_get_env_path():
    bash_env_path = (
        "/c/Users/David/Source/ducktools-pytui/.venv/Scripts:"
        "/mingw64/bin:"
        "/usr/bin:"
        "/c/Users/David/.local/bin:"
        "/usr/bin:"
    )

    with patch("subprocess.run") as run_mock:
        run_mock.return_value.stdout = bash_env_path
        env_path = bash.GitBashShell(BASH_PATH).get_env_path()

        assert env_path == bash_env_path
        run_mock.assert_called_once_with(
            [BASH_PATH, "-ic", "echo $PATH"],
            text=True,
            capture_output=True,
        )


def test_get_venv_shell_command():
    venv = "/c/Users/David/Source/ducktools-pytui/.venv"
    bash_path = (
        "/c/Users/David/Source/ducktools-pytui/.venv/Scripts:"
        "/mingw64/bin:"
        "/usr/bin:"
        "/c/Users/David/.local/bin"
    )
    prompt = "pytui: .venv"

    env = {
        "PYTUI_PATH": bash_path,
        "PYTUI_VIRTUAL_ENV": venv,
        "PYTUI_VIRTUAL_ENV_PROMPT": prompt,
    }

    with patch("ducktools.pytui.shells.bash.get_shell_script") as script_mock:
        def identity(x): return x

        script_mock.side_effect = identity

        shell = bash.GitBashShell(BASH_PATH)

        cmd, env_updates = shell.get_venv_shell_command(env)

        assert cmd == [BASH_PATH, "--rcfile", "activate_pytui.sh"]
        assert env_updates == {}  # environment variables are set in the script



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import functools
import json
import sys
import subprocess

from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from ducktools.pythonfinder import PythonInstall
from ducktools.pytui.runtime_installers import uv
from ducktools.pytui.runtime_installers.uv import UVPythonListing, UVManager

if sys.platform == "win32":
    example_folder = Path(__file__).parent / "example_data" / "win32"
else:
    example_folder = Path(__file__).parent / "example_data"


@functools.lru_cache(maxsize=None)
def uv_download_json() -> str:
    return (example_folder / "uv_download_list.json").read_text()


@functools.lru_cache(maxsize=None)
def uv_install_json() -> str:
    return (example_folder / "uv_install_list.json").read_text()


@pytest.fixture()
def uv_installed_pythons(uv_python_dir) -> list[uv.UVPythonListing]:
    manager = UVManager()
    pys = [uv.UVPythonListing.from_dict(manager=manager, entry=v) for v in json.loads(uv_install_json())]
    return pys


# noinspection PyPropertyAccess
def test_check_uv():
    manager = UVManager()
    with patch("shutil.which") as fake_which:
        fake_which.return_value = "~/.local/bin/uv"

        out = manager.executable

        fake_which.assert_called()
        assert out == "~/.local/bin/uv"

    del manager.executable

    with patch("shutil.which") as fake_which:
        fake_which.return_value = None

        out = manager.executable
        fake_which.assert_called()
        assert out is None


def test_fake_uv(uv_executable):
    manager = UVManager()
    assert manager.executable == "uv"


# noinspection PyPropertyAccess
def test_uv_python_dir():
    if sys.platform == "win32":
        base = "C:\\Users\\ducks\\AppData\\Roaming\\uv\\python"
    else:
        base = "/home/david/.local/share/uv/python"

    manager = UVManager()

    with patch("subprocess.run") as fake_process:
        cmd_output = MagicMock()
        cmd_output.stdout = f"{base}\n"
        fake_process.return_value = cmd_output

        out = manager.runtime_folder
        fake_process.assert_called()
        assert out == base

    del manager.runtime_folder

    with patch("subprocess.run", side_effect=FileNotFoundError) as fake_process:
        out = manager.runtime_folder
        fake_process.assert_called()
        assert out is None

    del manager.runtime_folder

    with patch(
        "subprocess.run",
        side_effect=subprocess.CalledProcessError(1, ""),
    ) as fake_process:
        out = manager.runtime_folder
        fake_process.assert_called()
        assert out is None


def test_fetch_installed(uv_executable, uv_python_dir):
    with patch("subprocess.run") as fake_process:
        cmd_output = MagicMock()
        cmd_output.stdout = uv_install_json()
        fake_process.return_value = cmd_output

        manager = UVManager()

        installed_pys = manager.fetch_installed()

        fake_process.assert_called_once_with(
            [
                "uv",
                "python",
                "list",
                "--output-format",
                "json",
                "--only-installed",
                "--python-preference",
                "only-managed",
                "--all-versions",
            ],
            capture_output=True,
            text=True,
            check=True,
        )

    if sys.platform == "win32":
        expected = [
            "cpython-3.14.0a5-windows-x86_64-none",
            "cpython-3.14.0a4-windows-x86_64-none",
            "cpython-3.13.2-windows-x86_64-none",
            "cpython-3.12.9-windows-x86_64-none",
            "cpython-3.11.11-windows-x86_64-none",
            "cpython-3.10.16-windows-x86_64-none",
            "cpython-3.9.21-windows-x86_64-none",
            "cpython-3.8.20-windows-x86_64-none",
            "pypy-3.11.11-windows-x86_64-none",
            # Note this final entry is *not* the key the list command reports
            # But is actually the key needed to uninstall correctly
            "pypy-3.10.19-windows-x86_64-none",
        ]
    else:
        expected = [
            "cpython-3.14.0a5-linux-x86_64-gnu",
            "cpython-3.14.0a4-linux-x86_64-gnu",
            "cpython-3.13.2-linux-x86_64-gnu",
            "cpython-3.12.9-linux-x86_64-gnu",
            "cpython-3.11.11-linux-x86_64-gnu",
            "cpython-3.10.16-linux-x86_64-gnu",
            "cpython-3.9.21-linux-x86_64-gnu",
            "cpython-3.8.20-linux-x86_64-gnu",
            "pypy-3.11.11-linux-x86_64-gnu",
            "pypy-3.10.19-linux-x86_64-gnu",
        ]

    assert [p.key for p in installed_pys] == expected


def test_fetch_downloads(uv_executable, uv_python_dir, uv_installed_pythons):
    with patch("subprocess.run") as fake_process, \
        patch.object(UVManager, "fetch_installed") as fake_installed:

        fake_installed.return_value = uv_installed_pythons

        manager = UVManager()

        cmd_output = MagicMock()
        cmd_output.stdout = uv_download_json()
        fake_process.return_value = cmd_output

        download_pys = manager.fetch_downloads()

        fake_installed.assert_called()
        fake_process.assert_called_once_with(
            [
                "uv",
                "python",
                "list",
                "--output-format",
                "json",
                "--only-downloads",
            ],
            capture_output=True,
            text=True,
            check=True,
        )

    if sys.platform == "win32":
        expected = [
            "cpython-3.14.0a5+freethreaded-windows-x86_64-none",
            "cpython-3.13.2+freethreaded-windows-x86_64-none",
            "cpython-3.7.9-windows-x86_64-none",
            "pypy-3.9.19-windows-x86_64-none",
            "pypy-3.8.16-windows-x86_64-none",
            "pypy-3.7.13-windows-x86_64-none",
        ]

    else:
        expected = [
            "cpython-3.14.0a5+freethreaded-linux-x86_64-gnu",
            "cpython-3.13.2+freethreaded-linux-x86_64-gnu",
            "cpython-3.7.9-linux-x86_64-gnu",
            "pypy-3.9.19-linux-x86_64-gnu",
            "pypy-3.8.16-linux-x86_64-gnu",
            "pypy-3.7.13-linux-x86_64-gnu",
        ]

    assert [p.key for p in download_pys] == expected


@pytest.mark.skipif(sys.platform != "win32", reason="Windows version of test")
def test_find_matching_listing_win(uv_installed_pythons):
    inst_313 = PythonInstall(
        version=(3, 13, 2, "final", 0),
        executable="C:\\Users\\ducks\\AppData\\Roaming\\uv\\python\\cpython-3.13.2-windows-x86_64-none\\python.exe",
        architecture="64bit",
        implementation="cpython",
        managed_by="Astral UV",
        metadata={"freethreaded": False},
        shadowed=False,
    )

    # Due to a metadata error, UV installs/uninstalls this as 3.10.19, but lists it as 3.10.16
    # 3.10.19 is the key we need, so make sure we still get the useful key even if it isn't the right version.
    pypy_310 = PythonInstall(
        version=(3, 10, 16, "final", 0),
        executable="C:\\Users\\ducks\\AppData\\Roaming\\uv\\python\\pypy-3.10.19-windows-x86_64-none\\python.exe",
        architecture="64bit",
        implementation="pypy",
        managed_by="Astral UV",
        metadata={"pypy_version": (7, 3, 19, "final", 0)},
        shadowed=False,
    )

    windows_install = PythonInstall(
        version=(3, 13, 1, "final", 0),
        executable="C:\\Users\\ducks\\AppData\\Local\\Programs\\Python\\Python313\\python.exe",
        architecture="64bit",
        implementation="cpython",
        managed_by="PythonCore",
        metadata={
            "Company": "PythonCore",
            "CompanyDisplayName": "Python Software Foundation",
            "CompanySupportUrl": "https://www.python.org/",
            "Tag": "3.13",
            "DisplayName": "Python 3.13 (64-bit)",
            "SupportUrl": "https://www.python.org/",
            "Version": "3.13.1",
            "SysVersion": "3.13",
            "SysArchitecture": "64bit",
            "InWindowsRegistry": True,
        },
        shadowed=False,
    )

    with patch.object(UVManager, "fetch_installed") as fake_installed:
        fake_installed.return_value = uv_installed_pythons

        manager = UVManager()

        uv_313 = manager.find_matching_listing(inst_313)

        assert uv_313.key == "cpython-3.13.2-windows-x86_64-none"

        uv_pypy_310 = manager.find_matching_listing(pypy_310)

        assert uv_pypy_310.key == "pypy-3.10.19-windows-x86_64-none"

        non_uv = manager.find_matching_listing(windows_install)

        assert non_uv is None


@pytest.mark.skipif(sys.platform == "win32", reason="Non-windows version of test")
def test_find_matching_listing_nonwin(uv_installed_pythons):
    inst_313 = PythonInstall(
        version=(3, 13, 2, "final", 0),
        executable="/home/david/.local/share/uv/python/cpython-3.13.2-linux-x86_64-gnu/bin/python",
        architecture="64bit",
        implementation="cpython",
        managed_by="Astral UV",
        metadata={"freethreaded": False},
        shadowed=False,
    )

    # Due to a metadata error, UV installs/uninstalls this as 3.10.19, but lists it as 3.10.16
    # 3.10.19 is the key we need, so make sure we still get the useful key even if it isn't the right version.
    pypy_310 = PythonInstall(
        version=(3, 10, 16, "final", 0),
        executable="/home/david/.local/share/uv/python/pypy-3.10.19-linux-x86_64-gnu/bin/python",
        architecture="64bit",
        implementation="pypy",
        managed_by="Astral UV",
        metadata={"pypy_version": (7, 3, 18, "final", 0)},
        shadowed=False,
    )

    pyenv_install = PythonInstall(
        version=(3, 13, 2, "final", 0),
        executable="/home/david/.pyenv/versions/3.13.2/bin/python",
        architecture="64bit",
        implementation="cpython",
        managed_by="pyenv",
        metadata={"freethreaded": False},
        shadowed=False,
    )

    with patch.object(UVManager, "fetch_installed") as fake_installed:
        fake_installed.return_value = uv_installed_pythons

        manager = UVManager()

        uv_313 = manager.find_matching_listing(inst_313)

        assert uv_313.key == "cpython-3.13.2-linux-x86_64-gnu"

        uv_pypy_310 = manager.find_matching_listing(pypy_310)

        assert uv_pypy_310.key == "pypy-3.10.19-linux-x86_64-gnu"

        non_uv = manager.find_matching_listing(pyenv_install)

        assert non_uv is None


def test_install(uv_executable, uv_python_dir):
    manager = UVManager()
    listing = UVPythonListing(
        manager=manager,
        key="cpython-3.14.0a5+freethreaded-windows-x86_64-none",
        version="3.14.0a5",
        version_parts={"major": 3, "minor": 14, "patch": 0},
        path=None,
        symlink=None,
        url="https://github.com/astral-sh/python-build-standalone/releases/download/20250212/cpython-3.14.0a5%2B20250212-x86_64-pc-windows-msvc-freethreaded%2Bpgo-full.tar.zst",
        os="windows",
        variant="freethreaded",
        implementation="cpython",
        arch="x86_64",
        libc="none",
    )
    with patch("subprocess.run") as fake_process:
        fake_out = MagicMock()
        fake_process.return_value = fake_out

        output = listing.install()

        fake_process.assert_called_once_with(
            [
                "uv",
                "python",
                "install",
                listing.key,
                "--color",
                "never",
                "--no-progress",
            ],
            capture_output=True,
            text=True,
            check=True,
        )

        assert output == fake_out


def test_uninstall(uv_executable, uv_python_dir):
    if sys.platform == "win32":
        py_path = "C:\\Users\\ducks\\AppData\\Roaming\\uv\\python\\cpython-3.13.2-windows-x86_64-none\\python.exe"
    else:
        # Yes this install path makes no sense
        # But it simplifies the amount of platform specific work needed
        py_path = "/home/david/.local/share/uv/python/cpython-3.13.2-windows-x86_64-none/bin/python3.13"

    manager = UVManager()

    listing = UVPythonListing(
        manager=manager,
        key="cpython-3.13.2-windows-x86_64-none",
        version="3.13.2",
        version_parts={"major": 3, "minor": 13, "patch": 2},
        path=py_path,
        symlink=None,
        url=None,
        os="windows",
        variant="default",
        implementation="cpython",
        arch="x86_64",
        libc="none",
    )

    with patch("subprocess.run") as fake_process, \
            patch("os.path.exists") as fake_exists:
        fake_out = MagicMock()
        fake_process.return_value = fake_out
        fake_exists.return_value = True

        output = listing.uninstall()

        fake_process.assert_called_once_with(
            [
                "uv",
                "python",
                "uninstall",
                listing.key,
                "--color",
                "never",
                "--no-progress",
            ],
            capture_output=True,
            text=True,
            check=True,
        )

        assert output == fake_out



from ducktools.lazyimporter import LazyImporter
from ducktools.lazyimporter.capture import capture_imports


laz = LazyImporter()

# These are actually imported lazily and do not exist within this module
with capture_imports(laz):
    import importlib_resources as resources  # noqa: F401



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import functools
import json
import os.path
import shutil
import subprocess
from pathlib import Path
from typing import ClassVar


from ducktools.classbuilder.prefab import get_attributes, prefab

from .base import RuntimeManager, PythonListing


class UVManager(RuntimeManager):
    organisation: ClassVar[str] = "Astral"

    @functools.cached_property
    def executable(self) -> str | None:
        return shutil.which("uv")

    @functools.cached_property
    def runtime_folder(self) -> str | None:
        try:
            cmd = subprocess.run(
                ["uv", "python", "dir"],
                check=True,
                capture_output=True,
                text=True,
            )
        except (FileNotFoundError, subprocess.CalledProcessError):
            py_dir = None
        else:
            py_dir = cmd.stdout.strip()
        return py_dir

    def fetch_installed(self):
        """
        Fetch Python installs managed by UV
        """
        installed_list_cmd = subprocess.run(
            [
                self.executable, "python", "list",
                "--output-format", "json",
                "--only-installed",
                "--python-preference", "only-managed",
                "--all-versions",
            ],
            capture_output=True,
            text=True,
            check=True,
        )
        json_data = json.loads(installed_list_cmd.stdout)
        installed_pys = [
            UVPythonListing.from_dict(manager=self, entry=v) for v in json_data
        ]

        return installed_pys

    @functools.lru_cache(maxsize=None)
    def _get_download_cache(self, all_versions=False):
        cmd = [
            self.executable, "python", "list",
            "--output-format", "json",
            "--only-downloads",
        ]
        if all_versions:
            cmd.append("--all-versions")

        download_list_cmd = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
        )
        full_download_list = json.loads(download_list_cmd.stdout)

        downloads = [
            UVPythonListing.from_dict(manager=self, entry=v)
            for v in full_download_list
        ]
        return downloads

    def fetch_downloads(self, all_versions=False):
        """
        Get available UV downloads and filter out any installs that are already present.

        :param all_versions: Include *ALL* possible installs
        :return: list of possible python installs
        """
        downloads = self._get_download_cache(all_versions=all_versions)

        installed_keys = {v.key for v in self.fetch_installed()}

        download_listings = self.sort_listings(
            v for v in downloads
            if v.key not in installed_keys
        )

        return download_listings


@prefab(kw_only=True)
class UVPythonListing(PythonListing):
    manager: UVManager

    # These extra parts are UV Specific
    version_parts: dict
    symlink: str | None
    os: str
    libc: str | None  # Apparently this is the string "none" instead of an actual None.

    def __prefab_post_init__(self, key: str, path: str | None) -> None:
        if path is None:
            self.key = key
            self.path = path
        else:
            # Resolve path always, sometimes UV gives a relative path to cwd.
            self.path = os.path.abspath(path)

            # UV bug - key and path can mismatch if someone typoed the metadata
            base_path = self.manager.runtime_folder
            if base_path:
                key_path = str(Path(self.path).relative_to(base_path).parts[0])
                self.key = key if key == key_path else key_path

    @classmethod
    def from_dict(cls, manager: UVManager, entry: dict):
        # designed to not fail if extra keys are added
        attrib_names = set(get_attributes(cls))

        kwargs = entry.copy()
        for key in entry.keys():
            if key not in attrib_names:
                del kwargs[key]

        return cls(manager=manager, **kwargs)

    def install(self):
        if self.path:
            # Can't install already installed Python
            return None

        cmd = [
            self.manager.executable, "python", "install",
            self.key,
            "--color", "never",
            "--no-progress",
        ]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
        )
        return result

    def uninstall(self):
        if not (self.path and os.path.exists(self.path)):
            # Can't uninstall non-installed Python
            return None

        cmd = [
            self.manager.executable, "python", "uninstall",
            self.key,
            "--color", "never",
            "--no-progress",
        ]
        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=True,
        )
        return result



# ducktools-pytui
# MIT License
#
# Copyright (c) 2025 David C Ellis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import functools
import json
from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from ducktools.pytui.runtime_installers.pythoncore import PythonCoreManager, PythonCoreListing

example_folder = Path(__file__).parents[1] / "example_data" / "win32"

@functools.lru_cache(maxsize=None)
def pymanager_download_json() -> str:
    return (example_folder / "pymanager_download_list.json").read_text()

@functools.lru_cache(maxsize=None)
def pymanager_install_json() -> str:
    return (example_folder / "pymanager_install_list.json").read_text()

@pytest.fixture()
def pymanager_installed_pythons():
    manager = PythonCoreManager()
    pys = [
        PythonCoreListing.from_dict(manager=manager, entry=v)
        for v in json.loads(pymanager_install_json()).get("versions", [])
    ]
    return pys


# noinspection PyPropertyAccess
def test_check_pymanager():
    manager = PythonCoreManager()
    with patch("shutil.which") as fake_which:
        fake_which.return_value = r"C:\Users\david\AppData\Local\Microsoft\WindowsApps\pymanager.exe"

        out = manager.executable

        fake_which.assert_called()
        assert out == r"C:\Users\david\AppData\Local\Microsoft\WindowsApps\pymanager.exe"

    del manager.executable

    with patch("shutil.which") as fake_which:
        fake_which.return_value = None

        out = manager.executable
        fake_which.assert_called()
        assert out is None

def test_fake_pymanager(pymanager_executable):
    manager = PythonCoreManager()
    assert manager.executable == "pymanager.exe"


def test_fetch_installed(pymanager_executable):
    with patch("subprocess.run") as fake_process:
        cmd_output = MagicMock()
        cmd_output.stdout = pymanager_install_json()
        fake_process.return_value = cmd_output

        manager = PythonCoreManager()

        installed_pys = manager.fetch_installed()

        fake_process.assert_called_once_with(
            [
                "pymanager.exe", "list", "--only-managed", "--format=json",
            ],
            capture_output=True,
            text=True,
            check=True,
        )

        expected = [
            'pythoncore-3.13t-64',
            'pythoncore-3.12-64',
            'pythoncore-3.12-32',
            'pythoncore-3.12-arm64',
            'pythoncore-3.14-64',
        ]

        assert [p.key for p in installed_pys] == expected


def test_fetch_downloads(pymanager_executable, pymanager_installed_pythons):
    # Fake as AMD64 for filtering
    with patch("platform.machine") as machine_mock, \
        patch("subprocess.run") as fake_process, \
        patch.object(PythonCoreManager, "fetch_installed") as fake_installed:

        machine_mock.return_value = "AMD64"

        fake_installed.return_value = pymanager_installed_pythons

        cmd_output = MagicMock()
        cmd_output.stdout = pymanager_download_json()
        fake_process.return_value = cmd_output

        manager = PythonCoreManager()
        download_pys = manager.fetch_downloads()

        fake_installed.assert_called()
        fake_process.assert_called_once_with(
            ["pymanager.exe", "list", "--online", "--format=json"],
            capture_output=True,
            text=True,
            check=True,
        )

        # Check values have been filtered
        for p in pymanager_installed_pythons:
            assert p.key not in download_pys

        expected = [
            "pythoncore-3.14t-64",
            "pythoncore-3.13-64",
            "pythoncore-3.11-64",
            "pythoncore-3.10-64",
            "pythoncore-3.9-64",
            "pythoncore-3.8-64",
            "pythoncore-3.7-64",
            "pythoncore-3.6-64",
            "pythoncore-3.5-64",
        ]

        assert [p.key for p in download_pys] == expected


def test_install(pymanager_executable):
    manager = PythonCoreManager()
    listing = PythonCoreListing(
        manager=manager,
        key='pythoncore-3.14-64',
        version='3.14.0b1',
        implementation='cpython',
        variant='default',
        arch='x86_64',
        path=None,
        url='https://www.python.org/ftp/python/3.14.0/python-3.14.0b1t-amd64.zip',
        name='Python 3.14.0b1',
        company="PythonCore",
        tag='3.14-dev-64',
        install_for=[
            "3.14.0b1-64",
            "3.14-64",
            "3-64",
            "3.14-dev-64",
            "3-dev-64",
        ]
    )

    with patch("subprocess.run") as fake_process, \
        patch.object(PythonCoreManager, "fetch_installed") as fake_installed:
        fake_installed.return_value = []
        listing.install()
        fake_process.assert_called_once_with(
            [
                "pymanager.exe", "install", "3.14.0b1-64", "-y"
            ],
            capture_output=True,
            text=True,
            check=True,
        )
